from fabrial import SequenceStep, StepRunner

from ..quince10GCE import Quince10GCEOven


async def create_oven(port: str, step: SequenceStep, runner: StepRunner) -> Quince10GCEOven:
    """Try to create a `Quince10GCEOven`, asking the user whether we should retry on failure."""
    while True:
        try:
            return Quince10GCEOven(port)
        except OSError:
            await runner.prompt_retry_cancel(step, "Failed to connect to oven.")



from abc import abstractmethod
from typing import Protocol


class Oven(Protocol):
    """
    Protocol for objects that represent ovens. This protocol does not care if the oven uses Celsius
    or Fahrenheit, just be consistent.
    """

    @abstractmethod
    def read_temperature(self) -> float | None:
        """
        Read the oven's temperature. This must free the oven (i.e. close the port) after reading the
        temperature.

        Returns
        -------
        The temperature if it could be read, `None` otherwise.
        """
        ...

    @abstractmethod
    def read_setpoint(self) -> float | None:
        """
        Read the oven's setpoint. This must free the oven (i.e. close the port) after reading the
        setpoint.

        Returns
        -------
        The setpoint if it could be read, `None` otherwise.
        """
        ...

    @abstractmethod
    def set_setpoint(self, setpoint: float) -> bool:
        """
        Set the oven's setpoint. This must free the oven (i.e. close the port) after setting the
        setpoint.

        Returns
        -------
        Whether the operation succeeded.
        """
        ...

    @abstractmethod
    def minimum_setpoint(self) -> float:
        """Get the minimum setpoint."""
        ...

    @abstractmethod
    def maximum_setpoint(self) -> float:
        """Get the maximum setpoint."""
        ...



from pathlib import Path

from fabrial import FilesDescription, ItemWidget, Substitutions
from PyQt6.QtGui import QIcon

from ...constants import (
    INCREMENT_LABEL,
    INTERVAL_LABEL,
    MINIMUM_MEASUREMENTS_LABEL,
    OVEN_PORT_LABEL,
    TEMPERATURES_FILENAME,
    TOLERANCE_LABEL,
)
from ...quince10GCE import MAXIMUM_SETPOINT
from ...widgets import OvenStabilizationWidget

BASE_NAME = "Increment Setpoint"
DIRECTORY = Path(__file__).parent
ICON = QIcon(str(DIRECTORY.joinpath("thermometer--plus.png")))
DESCRIPTIONS_DIRECTORY = DIRECTORY.joinpath("descriptions")


class IncrementSetpointWidget(ItemWidget):
    """Increment the oven's setpoint and stabilize; widget."""

    def __init__(
        self,
        port: str | None,
        increment: float,
        measurement_interval_ms: int,
        minimum_measurements: int,
        tolerance: float,
    ):
        self.data_widget = OvenStabilizationWidget(
            INCREMENT_LABEL,
            -MAXIMUM_SETPOINT,
            MAXIMUM_SETPOINT,
            port,
            increment,
            measurement_interval_ms,
            minimum_measurements,
            tolerance,
        )

        ItemWidget.__init__(
            self,
            self.data_widget,
            BASE_NAME,
            ICON,
            FilesDescription(
                DESCRIPTIONS_DIRECTORY,
                BASE_NAME,
                Substitutions(
                    parameters={
                        "OVEN_PORT": OVEN_PORT_LABEL,
                        "INCREMENT": INCREMENT_LABEL,
                        "INTERVAL": INTERVAL_LABEL,
                        "MINIMUM_MEASUREMENTS": MINIMUM_MEASUREMENTS_LABEL,
                        "TOLERANCE": TOLERANCE_LABEL,
                    },
                    data_recording={"TEMPERATURE_FILE": TEMPERATURES_FILENAME},
                ),
            ),
        )

        self.data_widget.temperature_spinbox.editingFinished.connect(self.handle_increment_change)

    def handle_increment_change(self):
        """Handle the value of the increment spinbox changing."""
        self.setWindowTitle(f"{BASE_NAME} ({self.data_widget.temperature_spinbox.value()} degrees)")



from pathlib import Path
from typing import Any

from fabrial import SequenceStep, StepRunner

from ...utility import oven as oven_util, sequence as sequence_util
from .no_stabilize_widget import BASE_NAME


class NoStabilizeStep(SequenceStep):
    """Set the oven's setpoint without waiting for the temperature to stabilize; step."""

    def __init__(self, port: str, setpoint: float):
        self.port = port
        self.setpoint = setpoint

    async def run(self, runner: StepRunner, data_directory: Path):  # implementation
        oven = await oven_util.create_oven(self.port, self, runner)
        # clamp the setpoint
        setpoint = await sequence_util.clamp_setpoint(self.setpoint, oven, self, runner)
        # set the setpoint
        await sequence_util.set_setpoint_check(setpoint, oven, self, runner)

    def reset(self):  # implementation
        return  # nothing to do here

    def name(self) -> str:
        return BASE_NAME

    def metadata(self) -> dict[str, Any]:
        return {"Selected Oven Port": self.port, "Selected Setpoint": self.setpoint}



import csv
import math
import time
from contextlib import AsyncExitStack
from datetime import datetime
from pathlib import Path
from typing import TextIO

from fabrial import SequenceStep, StepCancellation, StepRunner
from fabrial.plotting import LineHandle, PlotSettings, SymbolParams

from ..oven import Oven

MAXIMUM_ALLOWED_DISCONNECTS = 5


# public
class StabilizeTask:
    """
    Utility class that sets an oven's setpoint and waits for the temperature to stabilize.

    Parameters
    ----------
    oven
        The oven to use. It must already be connected.
    measurement_interval_ms
        The measurement interval in milliseconds.
    minimum_measurements
        The minimum number of within-**tolerance** measurements that need to be taken for the
        temperature to be "stable".
    tolerance
        How far off the temperature can be from the setpoint before the measurement is considered
        "unstable".
    step
        The step calling this function.
    runner
        The `StepRunner` running **step**.
    data_directory
        Where the step is recording data.
    temperature_filename
        The name of the file to record the temperatures to.

    Notes
    -----
    - The **oven** must already be connected, and the caller is responsible for disconnecting it.
    - This class will ensure the **setpoint** is within the oven's allowed range and will prompt
    the user if the setpoint is invalid.
    """

    def __init__(
        self,
        oven: Oven,
        measurement_interval_ms: int,
        minimum_measurements: int,
        tolerance: float,
        step: SequenceStep,
        runner: StepRunner,
        data_directory: Path,
        temperature_filename: str,
    ):
        self.oven = oven
        self.measurement_interval_ms = measurement_interval_ms
        self.minimum_measurements = minimum_measurements
        self.tolerance = tolerance
        self.step = step
        self.runner = runner
        self.data_directory = data_directory
        self.temperature_filename = temperature_filename

    # public
    async def run(self, selected_setpoint: float):
        """
        Set the oven's setpoint and wait for the temperature to stabilize. Record and plot the
        temperature while stabilizing.
        """
        # make sure the setpoint is within range
        setpoint = await clamp_setpoint(selected_setpoint, self.oven, self.step, self.runner)

        # set the setpoint
        await set_setpoint_check(setpoint, self.oven, self.step, self.runner)

        async with AsyncExitStack() as exit_stack:
            # more setup
            file = await self.create_temperature_file(exit_stack)
            # set up the plot
            plot_handle = await exit_stack.enter_async_context(
                self.runner.create_plot(
                    self.step,
                    "Oven",
                    PlotSettings("Oven Temperature", "Time (s)", "Temperature (°C)"),
                )
            )
            line_handle = await plot_handle.add_line(None, None, SymbolParams("o", "orange", 2))

            # how many "stable" measurements we've taken
            stable_measurement_count = 0
            # record the start time
            start_time = time.time()

            while stable_measurement_count < self.minimum_measurements:
                # read the temperature
                temperature, failure_count = await read_temperature_check(
                    self.oven, self.step, self.runner
                )
                # punish failed measurements by making stabilization take longer. Clamp to 0
                stable_measurement_count = max(stable_measurement_count - failure_count, 0)
                # record data
                await self.record_measurement(file, start_time, temperature, line_handle)
                # only temperatures within tolerance are "stable"
                if abs(setpoint - temperature) <= self.tolerance:
                    stable_measurement_count += 1
                else:  # otherwise we are unstable; reset
                    stable_measurement_count = 0

    async def record_measurement(
        self, file: TextIO, start_time: float, temperature: float, line_handle: LineHandle
    ):
        """
        Record the measurement datetime, temperature, and oven setpoint.

        Raises
        ------
        StepCancellation
            The user asked to cancel the step.
        """
        setpoint = await read_setpoint_check(self.oven, self.step, self.runner)
        csv.writer(file, lineterminator="\n").writerow([str(datetime.now()), temperature, setpoint])
        # add the datapoint to the line
        line_handle.add_point(time.time() - start_time, temperature)

    # private
    async def create_temperature_file(self, exit_stack: AsyncExitStack) -> TextIO:
        """
        Try to create the temperature file. If the operation fails, ask the user whether we should
        retry.

        Raises
        ------
        StepCancellation
            The user asked to cancel the step.
        """
        while True:
            try:
                file = exit_stack.enter_context(
                    open(self.data_directory.joinpath(self.temperature_filename), "w")
                )
                # write the header
                csv.writer(file, lineterminator="\n").writerow(
                    ["Datetime", "Temperature", "Setpoint"]
                )
                return file
            except OSError:
                await self.runner.prompt_retry_cancel(
                    self.step, "Failed to create temperatures file."
                )


# public
async def set_setpoint_check(setpoint: float, oven: Oven, step: SequenceStep, runner: StepRunner):
    """
    Try to set the oven's setpoint to **setpoint**. If the operation fails
    `MAXIMUM_ALLOWED_DISCONNECTS` times, ask the user if they want to retry or cancel the step.

    Raises
    ------
    StepCancellation
        The user asked to cancel the step.

    Notes
    -----
    This function *does not* clamp the setpoint to the oven's allowed range. You are responsible
    for ensuring the setpoint is valid.
    """
    while True:
        # try to set the setpoint
        disconnect_count = 0
        while not oven.set_setpoint(setpoint):
            disconnect_count += 1
            if disconnect_count < MAXIMUM_ALLOWED_DISCONNECTS:
                await step.sleep(1)  # arbitrary delay
                continue

            # we couldn't set the setpoint; ask the user what to do
            await runner.prompt_retry_cancel(step, "Failed to set the oven's setpoint.")
            disconnect_count = 0  # reset the disconnect count

        # make sure the setpoint was actually set
        measured_setpoint = await read_setpoint_check(oven, step, runner)
        if not math.isclose(measured_setpoint, setpoint):
            await runner.prompt_retry_cancel(step, "The oven's setpoint was not actually set.")
            continue

        return  # if we get here we successfully set the setpoint


# public
async def read_temperature_check(
    oven: Oven, step: SequenceStep, runner: StepRunner
) -> tuple[float, int]:
    """
    Read the **oven**'s temperature. If the read fails `MAXIMUM_ALLOWED_DISCONNECTS` times, ask
    the user try retry or cancel the **step**.

    Returns
    -------
    A tuple of (the temperature, how many times we failed to measure).

    Raises
    ------
    StepCancellation
        The user asked to cancel the step.
    """
    failure_count = 0
    disconnect_count = 0
    while (temperature := oven.read_temperature()) is None:
        failure_count += 1
        disconnect_count += 1
        if disconnect_count < MAXIMUM_ALLOWED_DISCONNECTS:
            await step.sleep(1)  # arbitrary delay
            continue

        # notify the user
        await runner.prompt_retry_cancel(step, "Failed to read oven temperature.")
        disconnect_count = 0

    return (temperature, failure_count)


# public
async def read_setpoint_check(oven: Oven, step: SequenceStep, runner: StepRunner) -> float:
    """
    Try to read the **oven**'s setpoint. If the read fails `MAXIMUM_ALLOWED_DISCONNECTS` times, ask
    the user try retry or cancel the **step**.

    Returns
    -------
    The setpoint.

    Raises
    ------
    StepCancellation
        The user asked to cancel the step.
    """
    disconnect_count = 0
    while (setpoint := oven.read_temperature()) is None:
        disconnect_count += 1
        if disconnect_count < MAXIMUM_ALLOWED_DISCONNECTS:
            await step.sleep(1)  # arbitrary delay
            continue

        # notify the user
        await runner.prompt_retry_cancel(step, "Failed to read oven setpoint.")
        disconnect_count = 0

    return setpoint


# public
async def clamp_setpoint(
    setpoint: float, oven: Oven, step: SequenceStep, runner: StepRunner
) -> float:
    """
    Check that the **setpoint** is within the **oven**'s allowed range. If not, prompt the user on
    whether the clamp the setpoint or cancel the step.

    Returns
    -------
    A valid setpoint (possibly the original).

    Raises
    ------
    StepCancellation
        The user asked to cancel the step.
    """
    if setpoint < (minimum := oven.minimum_setpoint()):
        await prompt_for_clamp(
            "The selected setpoint is below the oven's minimum."
            "\n\nClamp to the minimum or cancel this step?",
            step,
            runner,
        )
        return minimum
    elif setpoint > (maximum := oven.maximum_setpoint()):
        await prompt_for_clamp(
            "The selected setpoint is above the oven's maximum."
            "\n\nClamp to the maximum or cancel this step?",
            step,
            runner,
        )
        return maximum

    return setpoint  # if it's within range just return the original


# private
async def prompt_for_clamp(message: str, step: SequenceStep, runner: StepRunner):
    """
    Prompt the user whether to clamp the setpoint or cancel the step.

    Raises
    ------
    StepCancellation
        The user asked to cancel the step.
    """
    match (response := await runner.prompt_user(step, message, {0: "Clamp", 1: "Cancel Step"})):
        case 0:  # clamp
            return
        case 1:  # cancel the step
            raise StepCancellation
        case _:  # this should never run
            raise ValueError(f"Got {response} but expected 0 or 1")



from pathlib import Path
from typing import Any

from fabrial import SequenceStep, StepRunner

from ...constants import TEMPERATURES_FILENAME
from ...utility import oven as oven_util, sequence as sequence_util
from ...utility.sequence import StabilizeTask
from .increment_setpoint_widget import BASE_NAME


class IncrementSetpointStep(SequenceStep):
    """Increment the oven's setpoint and stabilize; step."""

    def __init__(
        self,
        port: str,
        increment: float,
        measurement_interval_ms: int,
        minimum_measurements: int,
        tolerance: float,
    ):
        self.port = port
        self.increment = increment
        self.measurement_interval_ms = measurement_interval_ms
        self.minimum_measurements = minimum_measurements
        self.tolerance = tolerance

    async def run(self, runner: StepRunner, data_directory: Path):  # implementation
        oven = await oven_util.create_oven(self.port, self, runner)
        # read the current setpoint and add the increment
        setpoint = await sequence_util.read_setpoint_check(oven, self, runner) + self.increment

        await StabilizeTask(
            oven,
            self.measurement_interval_ms,
            self.minimum_measurements,
            self.tolerance,
            self,
            runner,
            data_directory,
            TEMPERATURES_FILENAME,
        ).run(setpoint)

    def reset(self):  # implementation
        return  # nothing to do

    def name(self) -> str:
        return BASE_NAME

    def metadata(self) -> dict[str, Any]:
        return {
            "Selected Oven Port": self.port,
            "Selected Increment": self.increment,
            "Selected Measurement Interval (ms)": self.measurement_interval_ms,
            "Selected Minimum Measurements": self.minimum_measurements,
            "Selected Tolerance": self.tolerance,
        }



import typing
from typing import Mapping, Self

from fabrial import Json, SequenceStep, WidgetDataItem

from ...constants import OVEN_PORT, SETPOINT
from .no_stabilize_step import NoStabilizeStep
from .no_stabilize_widget import NoStabilizeWidget


class NoStabilizeItem(WidgetDataItem):
    """Set the oven's setpoint without waiting for the temperature to stabilize; item."""

    def __init__(self, port: str | None, setpoint: float):
        self.data_widget = NoStabilizeWidget(port, setpoint)

    @classmethod
    def deserialize(cls, serialized_obj: Mapping[str, Json]) -> Self:  # implementation
        item = cls(
            typing.cast(str, serialized_obj[OVEN_PORT]),
            typing.cast(float, serialized_obj[SETPOINT]),
        )
        item.data_widget.handle_setpoint_change()
        return item

    def serialize(self) -> dict[str, Json]:  # implementation
        return {
            OVEN_PORT: self.data_widget.port_combo_box.currentText(),
            SETPOINT: self.data_widget.setpoint_spinbox.value(),
        }

    def widget(self) -> NoStabilizeWidget:  # implementation
        return self.data_widget

    def create_sequence_step(self, _) -> SequenceStep:  # implementation
        return NoStabilizeStep(
            self.data_widget.port_combo_box.currentText(), self.data_widget.setpoint_spinbox.value()
        )



import typing
from typing import Mapping, Self

from fabrial import Json, SequenceStep, WidgetDataItem

from ...constants import OVEN_PORT
from .record_setpoint_step import RecordSetpointStep
from .record_setpoint_widget import RecordSetpointWidget


class RecordSetpointItem(WidgetDataItem):
    """Record the oven's setpoint once; item."""

    def __init__(self, port: str | None = None):
        self.data_widget = RecordSetpointWidget(port)

    @classmethod
    def deserialize(cls, serialized_obj: Mapping[str, Json]) -> Self:  # implementation
        return cls(typing.cast(str, serialized_obj[OVEN_PORT]))

    def serialize(self) -> dict[str, Json]:  # implementation
        return {OVEN_PORT: self.data_widget.port_combo_box.currentText()}

    def widget(self) -> RecordSetpointWidget:  # implementation
        return self.data_widget

    def create_sequence_step(self, _) -> SequenceStep:  # implementation
        return RecordSetpointStep(self.data_widget.port_combo_box.currentText())



from pathlib import Path

from fabrial import ItemWidget, TextDescription
from fabrial.custom_widgets import Widget
from PyQt6.QtGui import QIcon
from PyQt6.QtWidgets import QFormLayout

from ...constants import OVEN_PORT_LABEL
from ...widgets import PortComboBox

NAME = "Record Setpoint"
ICON = QIcon(str(Path(__file__).parent.joinpath("tag-label-green.png")))
FILENAME = "setpoint.csv"


class RecordSetpointWidget(ItemWidget):
    """Record the oven's setpoint once; widget."""

    def __init__(self, port: str | None):
        layout = QFormLayout()
        self.port_combo_box = PortComboBox(port)
        layout.addRow(OVEN_PORT_LABEL, self.port_combo_box)

        ItemWidget.__init__(
            self,
            Widget(layout),
            NAME,
            ICON,
            TextDescription(
                NAME,
                "Record the oven's setpoint once.",
                {OVEN_PORT_LABEL: "The COM port the oven is connected to."},
                {FILENAME: "The recorded setpoint."},
            ),
        )



# serialization keys
OVEN_PORT = "oven_port"
SETPOINT = "setpoint"
INCREMENT = "increment"
MEASUREMENT_INTERVAL = "measurement_interval"
MINIMUM_MEASUREMENTS = "minimum_measurements"
TOLERANCE = "tolerance"


# files
TEMPERATURES_FILENAME = "temperatures.csv"

# labels
OVEN_PORT_LABEL = "Oven Port"
SETPOINT_LABEL = "Setpoint"
INCREMENT_LABEL = "Increment"
INTERVAL_LABEL = "Measurement Interval"
MINIMUM_MEASUREMENTS_LABEL = "Minimum Measurements for Stability"
TOLERANCE_LABEL = "Stability Tolerance"



from minimalmodbus import Instrument

from .oven import Oven

# Celsius
MINIMUM_SETPOINT = 0.0
MAXIMUM_SETPOINT = 232.0

TEMPERATURE_REGISTER = 1
SETPOINT_REGISTER = 0
NUMBER_OF_DECIMALS = 1


class Quince10GCEOven(Oven):
    """
    Represents a physical [Quincy 10 GCE lab oven](https://quincylab.com/digital-lab-ovens).

    Parameters
    ----------
    port
        The port (i.e. "COM4") to use.

    Raises
    ------
    SerialException
        The port cannot be opened.
    """

    def __init__(self, port: str):
        # this instantly opens the serial port so it could throw an error
        # `close_port_after_each_call` ensures multiple instances can exist simultaneously
        self.instrument = Instrument(port, 1)

    def read_temperature(self) -> float | None:  # implementation
        try:
            return float(self.instrument.read_register(TEMPERATURE_REGISTER, NUMBER_OF_DECIMALS))
        except OSError:
            return None

    def read_setpoint(self) -> float | None:  # implementation
        try:
            return float(self.instrument.read_register(SETPOINT_REGISTER, NUMBER_OF_DECIMALS))
        except OSError:
            return None

    def set_setpoint(self, setpoint: float) -> bool:  # implementation
        try:
            self.instrument.write_register(SETPOINT_REGISTER, setpoint, NUMBER_OF_DECIMALS)
            return True
        except OSError:
            return False

    def minimum_setpoint(self) -> float:  # implementation
        return MINIMUM_SETPOINT

    def maximum_setpoint(self) -> float:  # implementation
        return MAXIMUM_SETPOINT



from serial.tools import list_ports as ports_util


def list_ports() -> list[str]:
    """
    Get a list of available ports.

    # Returns
    A list of available port strings (i.e. "COM4" on Windows or "/dev/ttyUSB0" on
    Linux).
    """
    return [port_info.device for port_info in ports_util.comports()]



from pathlib import Path

from fabrial import ItemWidget, TextDescription
from fabrial.custom_widgets import DoubleSpinBox, Widget
from PyQt6.QtGui import QIcon
from PyQt6.QtWidgets import QFormLayout

from ...constants import OVEN_PORT_LABEL, SETPOINT_LABEL
from ...quince10GCE import MAXIMUM_SETPOINT, MINIMUM_SETPOINT
from ...widgets import PortComboBox

BASE_NAME = "Set Setpoint No Stabilization"
ICON = QIcon(str(Path(__file__).parent.joinpath("thermometer.png")))


class NoStabilizeWidget(ItemWidget):
    """Set the oven's setpoint without waiting for the temperature to stabilize; widget."""

    def __init__(self, port: str | None, setpoint: float):
        layout = QFormLayout()
        self.port_combo_box = PortComboBox(port)
        self.setpoint_spinbox = DoubleSpinBox(2, MINIMUM_SETPOINT, MAXIMUM_SETPOINT, setpoint)
        layout.addRow(OVEN_PORT_LABEL, self.port_combo_box)
        layout.addRow(SETPOINT_LABEL, self.setpoint_spinbox)

        ItemWidget.__init__(
            self,
            Widget(layout),
            BASE_NAME,
            ICON,
            TextDescription(
                BASE_NAME,
                "Set the oven's setpoint without waiting for it to stabilize.",
                {
                    OVEN_PORT_LABEL: "The COM port the oven is connected to.",
                    SETPOINT_LABEL: "The setpont to set the oven to.",
                },
            ),
        )

        self.setpoint_spinbox.editingFinished.connect(self.handle_setpoint_change)

    def handle_setpoint_change(self):
        """Handle the value of the setpoint spinbox changing."""
        self.setWindowTitle(f"{BASE_NAME} ({self.setpoint_spinbox.value()} degrees)")



from pathlib import Path

from fabrial import SequenceStep, StepRunner

from ...utility import oven as oven_util, sequence as sequence_util
from .record_temperature_widget import FILENAME, NAME


class RecordTemperatureStep(SequenceStep):
    """Record the oven's temperature once; step."""

    def __init__(self, port: str):
        self.port = port

    async def run(self, runner: StepRunner, data_directory: Path):  # implementation
        oven = await oven_util.create_oven(self.port, self, runner)
        temperature = await sequence_util.read_temperature_check(oven, self, runner)

        with open(data_directory.joinpath(FILENAME), "w") as f:
            f.write(f"temperature\n{temperature}\n")

    def reset(self):  # implementation
        return  # nothing to do

    def name(self) -> str:  # implementation
        return NAME



from .increment_setpoint import IncrementSetpointItem
from .record_setpoint import RecordSetpointItem
from .record_temperature import RecordTemperatureItem
from .set_setpoint import SetSetpointItem
from .set_setpoint_no_stabilize import NoStabilizeItem



import typing
from typing import Mapping, Self

from fabrial import Json, SequenceStep, WidgetDataItem

from ...constants import MEASUREMENT_INTERVAL, MINIMUM_MEASUREMENTS, OVEN_PORT, SETPOINT, TOLERANCE
from .set_setpoint_step import SetSetpointStep
from .set_setpoint_widget import SetSetpointWidget


class SetSetpointItem(WidgetDataItem):
    """Set the oven's setpoint and stabilize; item."""

    def __init__(
        self,
        port: str | None,
        setpoint: float,
        measurement_interval_ms: int,
        minimum_measurements: int,
        tolerance: float,
    ):
        self.data_widget = SetSetpointWidget(
            port, setpoint, measurement_interval_ms, minimum_measurements, tolerance
        )

    @classmethod
    def deserialize(cls, serialized_obj: Mapping[str, Json]) -> Self:  # implementation
        item = cls(
            typing.cast(str, serialized_obj[OVEN_PORT]),
            typing.cast(float, serialized_obj[SETPOINT]),
            typing.cast(int, serialized_obj[MEASUREMENT_INTERVAL]),
            typing.cast(int, serialized_obj[MINIMUM_MEASUREMENTS]),
            typing.cast(float, serialized_obj[TOLERANCE]),
        )
        item.data_widget.handle_setpoint_change()
        return item

    def serialize(self) -> dict[str, Json]:  # implementation
        widget = self.data_widget.data_widget
        return {
            OVEN_PORT: widget.port_combo_box.currentText(),
            SETPOINT: widget.temperature_spinbox.value(),
            MEASUREMENT_INTERVAL: widget.interval_spinbox.value(),
            MINIMUM_MEASUREMENTS: widget.minimum_measurements_spinbox.value(),
            TOLERANCE: widget.tolerance_spinbox.value(),
        }

    def widget(self) -> SetSetpointWidget:  # implementation
        return self.data_widget

    def create_sequence_step(self, _) -> SequenceStep:  # implementation
        widget = self.data_widget.data_widget
        return SetSetpointStep(
            widget.port_combo_box.currentText(),
            widget.temperature_spinbox.value(),
            widget.interval_spinbox.value(),
            widget.minimum_measurements_spinbox.value(),
            widget.tolerance_spinbox.value(),
        )



from fabrial.custom_widgets import DoubleSpinBox, SpinBox, Widget
from PyQt6.QtGui import QShowEvent
from PyQt6.QtWidgets import QComboBox, QFormLayout

from .constants import INTERVAL_LABEL, MINIMUM_MEASUREMENTS_LABEL, OVEN_PORT_LABEL, TOLERANCE_LABEL
from .utility import ports


class PortComboBox(QComboBox):
    """A `QComboBox` for COM ports."""

    def __init__(self, initial_port: str | None = None):
        QComboBox.__init__(self)
        self.reload_ports()
        self.setCurrentText(initial_port)

    def reload_ports(self):
        """Refresh the list of ports."""
        current = self.currentText()
        self.clear()
        self.addItems(ports.list_ports())
        self.setCurrentText(current)  # restore the selection

    def showEvent(self, event: QShowEvent | None):
        # refresh the list of ports every time this widget is shown
        self.reload_ports()
        QComboBox.showEvent(self, event)


class OvenStabilizationWidget(Widget):
    """
    Contains entries for running a temperature stabilization step.

    Parameters
    ----------
    temperature_label
        The label to use for the temperature spinbox.
    minimum_temperature
        The minimum value for the temperature spinbox.
    maximum_temperature
        The maximum value for the temperature spinbox.
    port
        The initial port to select.
    temperature
        The initial value of the temperature spinbox.
    measurement_interval_ms
        The initial value of the measurement interval spinbox.
    minimum_measurements
        The initial value of the minimum measurements spinbox.
    tolerance
        The initial value of the tolerance spinbox.
    """

    def __init__(
        self,
        temperature_label: str,
        minimum_temperature: float,
        maximum_temperature: float,
        port: str | None,
        temperature: float,
        measurement_interval_ms: int,
        minimum_measurements: int,
        tolerance: float,
    ):
        layout = QFormLayout()
        Widget.__init__(self, layout)

        self.port_combo_box = PortComboBox(port)
        self.temperature_spinbox = DoubleSpinBox(
            2, minimum_temperature, maximum_temperature, temperature
        )
        self.interval_spinbox = SpinBox(10, initial_value=measurement_interval_ms)
        self.minimum_measurements_spinbox = SpinBox(2, initial_value=minimum_measurements)
        self.tolerance_spinbox = DoubleSpinBox(2, initial_value=tolerance)

        for label, widget in (
            (OVEN_PORT_LABEL, self.port_combo_box),
            (temperature_label, self.temperature_spinbox),
            (INTERVAL_LABEL, self.interval_spinbox),
            (MINIMUM_MEASUREMENTS_LABEL, self.minimum_measurements_spinbox),
            (TOLERANCE_LABEL, self.tolerance_spinbox),
        ):
            layout.addRow(label, widget)



from fabrial import PluginCategory

from .items import (
    IncrementSetpointItem,
    NoStabilizeItem,
    RecordSetpointItem,
    RecordTemperatureItem,
    SetSetpointItem,
)
from .oven import Oven


# Fabrial entry point
def categories() -> list[PluginCategory]:
    return [
        PluginCategory(
            "Quince 10 GCE Lab Oven",
            [
                SetSetpointItem(None, 0, 5000, 150, 1),
                IncrementSetpointItem(None, 10, 5000, 150, 1),
                NoStabilizeItem(None, 0),
                RecordTemperatureItem(),
                RecordSetpointItem(),
            ],
        ),
    ]
